# -*- coding: utf-8 -*-
# Copyright 2016 LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import requests


class RedOctoberException(EnvironmentError):
    """ This exception is raised from errors in the RedOctober Library. """

class RedOctoberRemoteException(RedOctoberException):
    """ This exception is raised to indicate issues returned from API. """

class RedOctoberDecryptException(RedOctoberRemoteException):
    """ This exception is raised when there are errors decrypting a file. """



# -*- coding: utf-8 -*-
# Copyright 2016 LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import requests

from datetime import timedelta

from .models.enum_user_role import EnumUserRole
from .models.enum_user_type import EnumUserType

from .exceptions import RedOctoberDecryptException
from .exceptions import RedOctoberRemoteException


class RedOctober(object):
    """ It provides Python bindings to a remote RedOctober server via HTTP(S).

    Additional documentation regarding the API endpoints is available at
    https://github.com/cloudflare/redoctober
    """

    def __init__(self, host, port, name, password, ssl=True):
        """ It initializes the RedOctober API using the provided credentials.

        Args:
            host (str): Host name/IP of Red October server.
            port (int): Port number of server.
            name (str): Account name for use as login.
            password (str): Password for account.
            ssl (bool): Is server SSL encrypted?
        """
        ssl = 'https' if ssl else 'http'
        self.uri_base = '%s://%s:%d' % (ssl, host, port)
        self.name = name
        self.password = password

    def create_vault(self):
        """ It creates a new vault.

        Create is the necessary first call to a new vault.
        It creates an admin account.

        Returns:
            bool: Status of vault creation
        """
        return self.call('create')

    def delegate(self, time=None, uses=None):
        """ It allows for the delegation of decryption rights.

        Delegate allows a user to delegate their decryption password to the
        server for a fixed period of time and for a fixed number of
        decryptions. If the user's account is not created, it creates it.
        Any new delegation overrides the previous delegation.

        Args:
            time (datetime.timedelta): Period of time that delegation is valid
                for.
            uses (int): Number of times that delegation can be used.
        Returns:
            bool: Status of delegation creation.
        """
        data = self._clean_mapping({
            'Time': time and '%ds' % time.total_seconds() or None,
            'Uses': uses,
        })
        return self.call('delegate', data=data)

    def create_user(self, user_type='rsa'):
        """ It creates a new user account.

        Allows an optional ``UserType`` to be specified which controls how the
        record is encrypted. This can have a value of either ``rsa`` or ``ecc``
        and if none is provided will default to ``rsa``.

        Args:
            user_type (str): Controls how the record is encrypted. This can have
                a value of either ``ecc`` or ``rsa``.
        Returns:
            bool: Status of user creation.
        """
        data = self._clean_mapping({
            'UserType': EnumUserType[user_type].name.upper(),
        })
        return self.call('create-user', data=data)

    def get_summary(self):
        """ It provides a list of keys and delegations for the server.

        Returns:
            dict: A mapping containing keys on the system, and users who have
                currently delegated their key to the server. Example:

                .. code-block:: python

                    {
                    "Live":{
                    "Bill":{"Admin":false,
                            "Type":"rsa",
                            "Expiry":"2013-11-26T08:42:29.65501032-08:00",
                            "Uses":3},
                    "Cat":{"Admin":false,
                           "Type":"rsa",
                           "Expiry":"2013-11-26T08:42:42.016311595-08:00",
                           "Uses":3},
                    "Dodo":{"Admin":false,
                            "Type":"rsa",
                            "Expiry":"2013-11-26T08:43:06.651429104-08:00",
                            "Uses":3}
                   },
                   "All":{
                    "Alice":{"Admin":true, "Type":"rsa"},
                    "Bill":{"Admin":false, "Type":"rsa"},
                    "Cat":{"Admin":false, "Type":"rsa"},
                    "Dodo":{"Admin":false, "Type":"rsa"}
                   }
        """
        return self.call('summary')

    def encrypt(self, minimum, owners, data):
        """ It allows a user to encrypt a piece of data.

        Args:
            minimum (int): Minimum number of users from ``owners`` set that
                must have delegated their keys to the server.
            owners (iter): Iterator of strings indicating users that may
                decrypt the document.
            data (str): Data to encrypt.
        Returns:
            str: Base64 encoded string representing the encrypted string.
        """
        data = self._clean_mapping({
            'Minimum': minimum,
            'Owners': owners,
            'Data': data.encode('base64'),
        })
        return self.call('encrypt', data=data)

    def decrypt(self, data):
        """ It allows a user to decrypt a piece of data.

        Args:
            data (str): Base64 encoded string representing the encrypted
                string.
        Raises:
            RedOctoberDecryptException: If not enough ``minimum`` users from
                the set of ``owners`` have delegated their keys to the server,
                or if the decryption credentials are incorrect.
        Returns:
            str: Decrypted string
        """
        data = self._clean_mapping({
            'Data': data,
        })
        try:
            return self.call('decrypt', data=data)
        except RedOctoberRemoteException as e:
            raise RedOctoberDecryptException(e.message)

    def get_owners(self, data):
        """ It provides the delegates required to decrypt a piece of data.

        Args:
            data (str): Base64 encoded string representing the encrypted
                string.
        Raises:
            RedOctoberDecryptException: If incorrect decryption credentials
                are provided.
        Returns:
            list: List of strings representing users that are able to decrypt
                the data.
        """
        data = self._clean_mapping({
            'Data': data,
        })
        try:
            return self.call('owners', data=data)
        except RedOctoberRemoteException as e:
            raise RedOctoberDecryptException(e.message)

    def change_password(self, new_password):
        """ It allows users to change their password.

        Args:
            name (str): Name of account.
            password (str): Password for account.
            new_password (str): New password for account.
        Returns:
            bool: Password change status.
        """
        data = self._clean_mapping({
            'NewPassword': new_password,
        })
        return self.call('password', data=data)

    def modify_user_role(self, modify_name, command='revoke'):
        """ It allows for administration of user roles.

        Args:
            modify_name (str): Name of account to modify.
            command (str): Command to apply to user:
                ``admin``: Promote user to administrator.
                ``revoke``: Revoke administrator rights.
                ``delete``: Delete user.
        Returns:
            bool: Role modfication status.
        """
        data = self._clean_mapping({
            'ToModify': modify_name,
            'Command': EnumUserRole[command].name,
        })
        return self.call('modify', data=data)

    def purge_delegates(self):
        """ It deletes all delegates for an encryption key.

        Returns:
            bool: Purge status.
        """
        return self.call('purge')

    def create_order(self, labels, duration, uses, data):
        """ It creates lets others users know delegations are needed.

        Args:
            labels (iter): Iterator of strings to label order with.
            duration (datetime.timedelta): Proposed duration of delegation.
            uses (int): Proposed delegation use amounts.
            data (str): Base64 encoded string representing the encrypted
                string.
        Returns:
            dict: Mapping representing the newly created order. Example:

            .. code-block:: python

                {
                    "Admins": [
                         "Bob",
                         "Eve"
                     ],
                     "AdminsDelegated": null,
                     "Delegated": 0,
                     "DurationRequested": 3.6e+12,
                     "Labels": [
                         "blue",
                         "red"
                     ],
                     "Name": "Alice",
                     "Num": "77da1cfd8962fb9685c15c84",
                     "TimeRequested": "2016-01-25T15:58:41.961906679-08:00",
                 }
        """
        data = self._clean_mapping({
            'Labels': labels,
            'Duration': '%ds' % duration.total_seconds(),
            'Uses': uses,
            'Data': data,
        })
        return self.call('order', data=data)

    def get_orders_outstanding(self):
        """ It returns a mapping of current orders.

        Returns:
            dict: Mapping representing the currently open orders. Example:

            .. code-block:: python

                {
                    "77da1cfd8962fb9685c15c84":{
                        "Name":"Alice",
                        "Num":"77da1cfd8962fb9685c15c84",
                        "TimeRequested":"2016-01-25T15:58:41.961906679-08:00",
                        "DurationRequested":3600000000000,
                        "Delegated":0,"
                        AdminsDelegated":null,
                        "Admins":["Bob, Eve"],
                        "Labels":["Blue","Red"]
                    }
                }
        """
        return self.call('orderout')

    def get_order_information(self, order_num):
        """ It gets information for a specified order.

        Args:
            order_num (str): Order number to get.
        Returns:
            dict: Mapping representing the order information. Example:

            .. code-block:: python

                {
                    "Admins": [
                        "Bob",
                        "Eve"
                    ],
                    "AdminsDelegated": null,
                    "Delegated": 0,
                    "DurationRequested": 3.6e+12,
                    "Labels": [
                        "blue",
                        "red"
                    ],
                    "Name": "Alice",
                    "Num": "77da1cfd8962fb9685c15c84",
                    "TimeRequested": "2016-01-25T15:58:41.961906679-08:00"
                }
        """
        data = self._clean_mapping({
            'OrderNum': order_num,
        })
        return self.call('orderinfo', data=data)

    def cancel_order(self, order_num):
        """ It cancels an order by number.

        Args:
            order_num (str): Order number to get.
        Returns:
            bool: Status of order cancellation.
        """
        data = self._clean_mapping({
            'OrderNum': order_num,
        })
        return self.call('ordercancel', data=data)

    def call(self, endpoint, method='POST', params=None, data=None):
        """ It calls the remote endpoint and returns the result, if success.

        Args:
            endpoint (str): RedOctober endpoint to call (e.g. ``newcert``).
            method (str): HTTP method to utilize for the Request.
            params: (dict|bytes) Data to be sent in the query string
                for the Request.
            data: (dict|bytes|file) Data to send in the body of the
                Request.
        Raises:
            RedOctoberRemoteException: In the event of a ``False`` in the
                ``success`` key of the API response.
        Returns:
            mixed: Data contained in ``result`` key of the API response, or
                ``True`` if there was no response data, but the call was a
                success.
        """
        endpoint = '%s/%s' % (self.uri_base, endpoint)
        if data:
            data.update({
                'Name': self.name,
                'Password': self.password,
            })
        response = requests.request(
            method=method,
            url=endpoint,
            params=params,
            data=data,
        )
        response = response.json()
        if response['Status'] != 'ok':
            raise RedOctoberRemoteException(
                '\n'.join([
                    'Response:',
                    '\n'.join(response.get('Response', [])),
                ])
            )
        try:
            return response['Response']
        except KeyError:
            return True

    def _clean_mapping(self, mapping):
        """ It removes false entries from mapping.

        Args:
            mapping (dict): Mapping to remove values from.
        Returns:
            dict: Mapping with no values evaluating to False.
        """
        return {k:v for k, v in mapping.iteritems() if v}



# -*- coding: utf-8 -*-
# Copyright 2016 LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from enum import Enum


class EnumUserType(Enum):
    """ It provides possible user encryption types for Red October.

    Attributes:
        rsa (int): Uses RSA encryption.
        ecc (int): Uses ECC encruyption.
    """
    rsa = 1
    ecc = 2



# -*- coding: utf-8 -*-
# Copyright 2016 LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import mock
import unittest

from datetime import timedelta

from ..models.enum_user_type import EnumUserType
from ..models.enum_user_role import EnumUserRole

from ..red_october import (RedOctober,
                           RedOctoberDecryptException,
                           RedOctoberRemoteException,
                           requests,
                           )


class TestRedOctober(unittest.TestCase):

    def setUp(self):
        super(TestRedOctober, self).setUp()
        self.name = 'name'
        self.password = 'password'
        self.uses = 987
        self.owners = ['Bob', 'Charles', 'Derphead', 'Ted']
        self.labels = ['private', 'super-secure', 'definitely-tweetable']
        self.delta = timedelta(seconds=1234)
        self.delta_str = '1234s'
        self.user_type = EnumUserType['ecc']
        self.data = 'DATA'
        self.user_role = EnumUserRole['delete']
        self.data64 = self.data.encode('base64')
        self.order_num = 'AnAwesomeIdentifier!!'
        self.red_october = RedOctober('test', 1, self.name, self.password)

    def test_uri_base_https(self):
        """ It should have an HTTP URI by default """
        self.assertIn('https://', self.red_october.uri_base)

    def test_uri_base_http(self):
        """ It should have an HTTP URI if someone decides to be crazy """
        red_october = RedOctober('test', 1, '', '', ssl=False)
        self.assertIn('http://', red_october.uri_base)

    @mock.patch.object(RedOctober, 'call')
    def test_create_vault(self, call):
        """ It should call with proper args """
        self.red_october.create_vault()
        call.assert_called_once_with('create')

    @mock.patch.object(RedOctober, 'call')
    def test_delegate(self, call):
        """ It should call with proper args """
        expect = {
            'Uses': self.uses,
            'Time': self.delta_str,
        }
        self.red_october.delegate(self.delta, self.uses)
        call.assert_called_once_with('delegate', data=expect)

    @mock.patch.object(RedOctober, 'call')
    def test_create_user(self, call):
        """ It should call with proper args """
        expect = {
            'UserType': self.user_type.name.upper(),
        }
        self.red_october.create_user(self.user_type.name)
        call.assert_called_once_with('create-user', data=expect)

    @mock.patch.object(RedOctober, 'call')
    def test_get_summary(self, call):
        """ It should call with proper args """
        self.red_october.get_summary()
        call.assert_called_once_with('summary')

    @mock.patch.object(RedOctober, 'call')
    def test_encrypt(self, call):
        """ It should call with proper args """
        expect = {
            'Minimum': self.uses,
            'Owners': self.owners,
            'Data': self.data64,
        }
        self.red_october.encrypt(self.uses, self.owners, self.data)
        call.assert_called_once_with('encrypt', data=expect)

    @mock.patch.object(RedOctober, 'call')
    def test_decrypt(self, call):
        """ It should call with proper args """
        expect = {
            'Data': self.data64,
        }
        self.red_october.decrypt(self.data64)
        call.assert_called_once_with('decrypt', data=expect)

    @mock.patch.object(RedOctober, 'call')
    def test_decrypt_fail(self, call):
        """ It should raise proper error on failure. """
        call.side_effect = RedOctoberRemoteException
        with self.assertRaises(RedOctoberDecryptException):
            self.red_october.decrypt(self.data64)

    @mock.patch.object(RedOctober, 'call')
    def test_get_owners(self, call):
        """ It should call with proper args """
        expect = {
            'Data': self.data64,
        }
        self.red_october.get_owners(self.data64)
        call.assert_called_once_with('owners', data=expect)

    @mock.patch.object(RedOctober, 'call')
    def test_get_owners_fail(self, call):
        """ It should raise proper error on failure. """
        call.side_effect = RedOctoberRemoteException
        with self.assertRaises(RedOctoberDecryptException):
            self.red_october.get_owners(self.data64)

    @mock.patch.object(RedOctober, 'call')
    def test_change_password(self, call):
        """ It should call with proper args """
        expect = {
            'NewPassword': 'new pass',
        }
        self.red_october.change_password('new pass')
        call.assert_called_once_with('password', data=expect)

    @mock.patch.object(RedOctober, 'call')
    def test_modify_user_role(self, call):
        """ It should call with proper args """
        expect = {
            'ToModify': 'Derphead',
            'Command': self.user_role.name,
        }
        self.red_october.modify_user_role('Derphead', self.user_role.name)
        call.assert_called_once_with('modify', data=expect)

    @mock.patch.object(RedOctober, 'call')
    def test_purge_delegates(self, call):
        """ It should call with proper args """
        self.red_october.purge_delegates()
        call.assert_called_once_with('purge')

    @mock.patch.object(RedOctober, 'call')
    def test_create_order(self, call):
        """ It should call with proper args """
        expect = {
            'Labels': self.labels,
            'Duration': self.delta_str,
            'Uses': self.uses,
            'Data': self.data64,
        }
        self.red_october.create_order(
            self.labels, self.delta, self.uses, self.data64,
        )
        call.assert_called_once_with('order', data=expect)

    @mock.patch.object(RedOctober, 'call')
    def test_get_orders_outstanding(self, call):
        """ It should call with proper args """
        self.red_october.get_orders_outstanding()
        call.assert_called_once_with('orderout')

    @mock.patch.object(RedOctober, 'call')
    def test_get_order_information(self, call):
        """ It should call with proper args """
        expect = {
            'OrderNum': self.order_num,
        }
        self.red_october.get_order_information(self.order_num)
        call.assert_called_once_with('orderinfo', data=expect)

    @mock.patch.object(RedOctober, 'call')
    def test_cancel_order(self, call):
        """ It should call with proper args """
        expect = {
            'OrderNum': self.order_num,
        }
        self.red_october.cancel_order(self.order_num)
        call.assert_called_once_with('ordercancel', data=expect)

    @mock.patch.object(requests, 'request')
    def test_call_request(self, requests):
        """ It should call requests with proper args """
        data = {'data': 'data',
                'Name': self.name,
                'Password': self.password,
                }
        requests_return = mock.MagicMock()
        requests.return_value = requests_return
        requests_return.json.return_value = {'Status': 'ok'}
        self.red_october.call('endpoint', 'method', 'params', data)
        requests.assert_called_once_with(
            method='method',
            url='https://test:1/endpoint',
            params='params',
            data=data,
        )

    @mock.patch.object(requests, 'request')
    def test_call_error(self, requests):
        """ It should raise on non-success response """
        requests().json.return_value = {'Status': 'not ok?'}
        with self.assertRaises(RedOctoberRemoteException):
            self.red_october.call('None')

    @mock.patch.object(requests, 'request')
    def test_call_success(self, requests):
        """ It should return result on success response """
        requests().json.return_value = {'Status': 'ok',
                                        'Response': ['result']}
        res = self.red_october.call(None)
        self.assertEqual(res, ['result'])

    @mock.patch.object(requests, 'request')
    def test_call_success_bool(self, requests):
        """ It should return True on success with no response """
        requests().json.return_value = {'Status': 'ok'}
        res = self.red_october.call(None)
        self.assertTrue(res)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright 2016 LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from enum import Enum


class EnumUserRole(Enum):
    """ It provides possible user roles for Red October.

    Attributes:
        delete (int): User is deleted or scheduled to be so.
        revoke (int): Normal user, revoke administrative rights.
        admin (int): User has administrative rights.
        user (int): Alias for `revoke`.
    """
    delete = 1
    revoke = 2
    admin = 3
    user = revoke



# -*- coding: utf-8 -*-
# Copyright 2016 LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

""" This library allows you to interact with a remote Red October Instance using Python.

Red October is a cryptographically-secure implementation of the two-person rule to
protect sensitive data. From a technical perspective, Red October is a software-based
encryption and decryption server. The server can be used to encrypt a payload in such
a way that no one individual can decrypt it. The encryption of the payload is
cryptographically tied to the credentials of the authorized users.

Authorized persons can delegate their credentials to the server for a period of time.
The server can decrypt any previously-encrypted payloads as long as the appropriate number
of people have delegated their credentials to the server.

This architecture allows Red October to act as a convenient decryption service. Other
systems, including CloudFlare’s build system, can use it for decryption and users can
delegate their credentials to the server via a simple web interface. All communication
with Red October is encrypted with TLS, ensuring that passwords are not sent in the clear.

* `Read more on the CloudFlare blog
  <https://blog.cloudflare.com/red-october-cloudflares-open-source-implementation-of-the-two-man-rule/>`_.
* `View the Red October source
  <https://github.com/cloudflare/redoctober>`_.
"""


from .red_october import RedOctober
